/*****************************************************************************/
/*
                             AuthCache.c


    THE GNU GENERAL PUBLIC LICENSE APPLIES DOUBLY TO ANYTHING TO DO WITH
                    AUTHENTICATION AND AUTHORIZATION!

    This package is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License, or any later
    version.

>   This package is distributed in the hope that it will be useful,
>   but WITHOUT ANY WARRANTY; without even the implied warranty of
>   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
>   GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


This modules maintains the authorization/authentication cache for WASD's
authorization environment.  A linked-list, binary tree (using the the
LIB$..._TREE routines) is used to store authentication records.  Records are
searched for and added using LIB$ library functions.

See AUTH.C for overall detail on the WASD authorization environment.


VERSION HISTORY
---------------
24-DEC-1999  MGD  break-in evasion period in report
28-AUG-1999  MGD  unbundled from AUTH.C for v6.1
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <libdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header files */
#include "wasd.h"

#define WASD_MODULE "AUTHCACHE"

#define FI_NOLI WASD_MODULE, __LINE__

/******************/
/* global storage */
/******************/

int  AuthDisplayRecordCount,
     AuthFreeRecordCount;

unsigned long  AuthCacheTreeBinTime [2];

struct AuthCacheRecordStruct  *AuthCacheLastHitPtr,
                              *AuthCacheTreeHead,
                              *AuthCacheTreeTailPtr;

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  WatchEnabled;
extern char ServerHostPort [];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;

/*****************************************************************************/
/*
Add the supplied authorization cache structure to the binary tree.
*/ 

int AuthCacheAddRecord (struct AuthCacheRecordStruct *RecordPtr)

{
   static int  NoDuplicatesFlag = 0;

   int  status;
   struct AuthCacheRecordStruct *InsertedRecordPtr;
   
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheAddRecord()\n");

   status = lib$insert_tree (&AuthCacheTreeHead,
                             RecordPtr,
                             &NoDuplicatesFlag,
                             &AuthCacheTreeCompareRecord,
                             &AuthCacheTreeAllocateRecord,
                             &InsertedRecordPtr, 0);
   if (Debug) fprintf (stdout, "lib$insert_tree() %%X%08.08X\n", status);

   return (status);
}

/*****************************************************************************/
/*
Search for the 'ThisRecordPtr' authorization cache structure in the binary
tree.  If found set the pointer pointed-to by 'FoundRecordPtrPtr' to point to
the appropriate record in the tree.  Improve repeated lookups by caching the
last successfully located record pointer.
*/ 

int AuthCacheFindRecord
(
struct AuthCacheRecordStruct *ThisRecordPtr,
struct AuthCacheRecordStruct **FoundRecordPtrPtr
)
{
   int  status;
   
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheFindRecord()\n");

   if (AuthCacheLastHitPtr != NULL)
   {
      if (!AuthCacheTreeCompareRecord (ThisRecordPtr, AuthCacheLastHitPtr, 0))
      {
         *FoundRecordPtrPtr = AuthCacheLastHitPtr;
         return (SS$_NORMAL);
      }
   }

   status = lib$lookup_tree (&AuthCacheTreeHead,
                             ThisRecordPtr,
                             &AuthCacheTreeCompareRecord,
                             FoundRecordPtrPtr);
   if (Debug) fprintf (stdout, "lib$lookup_tree() %%X%08.08X\n", status);

   if (VMSok (status))
   {
      AuthCacheLastHitPtr = *FoundRecordPtrPtr;
      return (status);
   }
   else
   {
      AuthCacheLastHitPtr = NULL;
      return (status);
   }
}

/*****************************************************************************/
/*
Called by lib$insert_tree() and lib$lookup_tree() in a variety of functions.
Returns negative number if node 1 is lexicographically less than the node 2,
positive if greater, or zero if exactly the same.
*/ 

int AuthCacheTreeCompareRecord
(
struct AuthCacheRecordStruct *acr1ptr,
struct AuthCacheRecordStruct *acr2ptr,
unsigned long Unused
)
{
   register int  cmp;
   
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeCompareRecord()\n");

   if (Debug)
      fprintf (stdout,
"%d %d\n\
%d %d\n\
%d %d\n\
%d |%s| %d |%s|\n\
%d |%s| %d |%s|\n\
%d |%s| %d |%s|\n\
%d |%s| %d |%s|\n\
%d |%s| %d |%s|\n",
      acr1ptr->SourceRealm, acr2ptr->SourceRealm,
      acr1ptr->SourceGroupWrite, acr2ptr->SourceGroupWrite,
      acr1ptr->SourceGroupRead, acr2ptr->SourceGroupRead,
      acr1ptr->RealmLength, acr1ptr->Realm,
      acr2ptr->RealmLength, acr2ptr->Realm,
      acr1ptr->AgentParameterLength, acr1ptr->AgentParameter,
      acr2ptr->AgentParameterLength, acr2ptr->AgentParameter,
      acr1ptr->GroupWriteLength, acr1ptr->GroupWrite,
      acr2ptr->GroupWriteLength, acr2ptr->GroupWrite,
      acr1ptr->GroupReadLength, acr1ptr->GroupRead,
      acr2ptr->GroupReadLength, acr2ptr->GroupRead,
      acr1ptr->UserNameLength, acr1ptr->UserName,
      acr2ptr->UserNameLength, acr2ptr->UserName);

   /* values of source indicators */
   if (cmp = acr1ptr->SourceRealm - acr2ptr->SourceRealm)
      return (cmp);
   if (cmp = acr1ptr->SourceGroupWrite - acr2ptr->SourceGroupWrite)
      return (cmp);
   if (cmp = acr1ptr->SourceGroupRead - acr2ptr->SourceGroupRead)
      return (cmp);

   /* lengths of all strings in the records */
   if (cmp = acr1ptr->RealmLength - acr2ptr->RealmLength)
      return (cmp);
   if (cmp = acr1ptr->AgentParameterLength - acr2ptr->AgentParameterLength)
      return (cmp);
   if (cmp = acr1ptr->GroupWriteLength - acr2ptr->GroupWriteLength)
      return (cmp);
   if (cmp = acr1ptr->GroupReadLength - acr2ptr->GroupReadLength)
      return (cmp);
   if (cmp = acr1ptr->UserNameLength - acr2ptr->UserNameLength)
      return (cmp);

   /* strings */
   if (cmp = strcmp (acr1ptr->Realm, acr2ptr->Realm))
      return (cmp);
   if (cmp = strcmp (acr1ptr->AgentParameter, acr2ptr->AgentParameter))
      return (cmp);
   if (cmp = strcmp (acr1ptr->GroupWrite, acr2ptr->GroupWrite))
      return (cmp);
   if (cmp = strcmp (acr1ptr->GroupRead, acr2ptr->GroupRead))
      return (cmp);
   return (strcmp (acr1ptr->UserName, acr2ptr->UserName));
}

/*****************************************************************************/
/*
Called by lib$insert_tree() in AuthDoAuthorization().  Insert the address
in 'acrptr' into the location pointed at by 'acrptrptr' (inside the binary tree
structure presumably!)
*/ 

AuthCacheTreeAllocateRecord
(
struct AuthCacheRecordStruct *acrptr,
struct AuthCacheRecordStruct **acrptrptr,
unsigned long Unused
)
{
   if (Debug) fprintf (stdout, "AuthCacheTreeAllocateRecord()\n");

   *acrptrptr = acrptr;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Clear all of the authentication records from the tree.  Will result in all
subsequent requests being re-authenticated from their respective on-disk
databases.  Called from the Admin.c and Control.c modules.
*/

int AuthCacheTreeFree ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeFree()\n");

   sys$gettim (&AuthCacheTreeBinTime);
   if (!AuthCacheTreeHead) return (SS$_NORMAL);
   AuthFreeRecordCount = 0;

   status = lib$traverse_tree (&AuthCacheTreeHead,
                               &AuthCacheTreeFreeRecord, 0);
   if (Debug) fprintf (stdout, "lib$traverse_tree() %%X%08.08X\n", status);

   if (AuthCacheTreeTailPtr) VmFree (AuthCacheTreeTailPtr);
   AuthCacheLastHitPtr = AuthCacheTreeHead = AuthCacheTreeTailPtr = NULL;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Free the dynamically allocated memory associated with the user tree record.
Called from AuthCacheTreeFree().
*/

AuthCacheTreeFreeRecord
(                      
struct AuthCacheRecordStruct *acrptr,
unsigned long  Unused
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeFreeRecord()\n");

   if (Debug)
      fprintf (stdout, "AuthCacheRecord |%s|%s|\n",
               acrptr->Realm, acrptr->UserName);

   AuthFreeRecordCount++;

   if (acrptr->LeftLink)
   {
      if (acrptr->LeftLink->UserDetailsPtr != NULL)
         VmFree (acrptr->LeftLink->UserDetailsPtr);
      if (acrptr->LeftLink->VmsUserProfilePtr != NULL)
         VmFree (acrptr->LeftLink->VmsUserProfilePtr);

      VmFree (acrptr->LeftLink);
   }
   else
      AuthCacheTreeTailPtr = acrptr;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Display all records in the authentication linked-list, binary tree.  Called
by the Control.c or the Request.c module.  For HTTP report, uses blocking
network write.
*/

AuthCacheTreeReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... User Authentication</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>User Authentication</H2>\n\
!AZ\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Since</TH></TR>\n\
<TR><TD>!AZ</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\\
<TR>\
<TH COLSPAN=2>Realm</TH>\
<TH COLSPAN=2>Group-R+W</TH>\
<TH>Group-R</TH>\
<TH COLSPAN=4>Agent-Param</TH>\
</TR>\n<TR>\
<TH>User</TH><TH>Access</TH>\
<TH COLSPAN=2>Scheme</TH>\
<TH>Most Recent</TH>\
<TH COLSPAN=4>Authenticated</TH>\
</TR>\n<TR>\
<TH COLSPAN=2></TH>\
<TH>Basic</TH><TH>Digest</TH>\
<TH></TH>\
<TH>Cache</TH><TH>Source</TH><TH>Fail</TH><TH>Break-in</TH>\
</TR>\n");

   static $DESCRIPTOR (NoneFaoDsc,
"<TR ALIGN=CENTER><TD COLSPAN=9><I>(none)</I></TD></TR>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</TABLE>\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;
   int  status;
   unsigned long  FaoVector [32];
   unsigned short  Length;
   char  Buffer [2048],
         TimeCurrent [32];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeReport()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = strcpy (TimeCurrent, DayDateTime (&rqptr->BinaryTime, 20));
   *vecptr++ = DayDateTime (&AuthCacheTreeBinTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$fao()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   AuthDisplayRecordCount = 0;
   status = lib$traverse_tree (&AuthCacheTreeHead,
                               &AuthCacheTreeReportRecord,
                               rqptr);
   if (Debug) fprintf (stdout, "lib$traverse_tree() %%X%08.08X\n", status);


   if (!AuthDisplayRecordCount)
      NetWriteBuffered (rqptr, 0, NoneFaoDsc.dsc$a_pointer,
                                 NoneFaoDsc.dsc$w_length);

   NetWriteBuffered (rqptr, NextTaskFunction,
                    EndPageFaoDsc.dsc$a_pointer,
                    EndPageFaoDsc.dsc$w_length);
}

/*****************************************************************************/
/*
Called by AuthCacheTreeReport().  Display the details of all records in the
linked-list binary tree.  Output the details via blocking network write.
*/

AuthCacheTreeReportRecord
(                      
struct AuthCacheRecordStruct *acrptr,
struct RequestStruct *rqptr
)
{
   static $DESCRIPTOR (ReportFaoDsc,
"<TR HEIGHT=1 COLSPAN=7><TD></TD>\
</TR>\n<TR>\
<TD COLSPAN=2>!AZ!AZ</TD>\
<TD COLSPAN=2>!AZ!AZ</TD>\
<TD>!AZ!AZ</TD>\
<TD COLSPAN=4>!AZ</TD>\
</TR>\n<TR>\
<TD>!AZ</TD><TD>!AZ!AZ!AZ</TD>\
<TD ALIGN=right>!UL</TD><TD ALIGN=right>!UL</TD>\
<TD><FONT SIZE=-1>!AZ</FONT></TD>\
<TD ALIGN=right>!UL</TD><TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL</TD><TD ALIGN=right>!AZ!UL!AZ</TD>\
</TR>\n");

   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  *UserCanStringPtr;
   char  Buffer [1024];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeReportRecord()\n");

   AuthDisplayRecordCount++;

   if ((UserCanStringPtr =
       AuthCanString (acrptr->AuthUserCan, AUTH_CAN_FORMAT_HTML)) == NULL)
      return (STS$K_ERROR);

   vecptr = FaoVector;

   *vecptr++ = acrptr->Realm;
   *vecptr++ = AuthSourceString (acrptr->Realm, acrptr->SourceRealm);

   *vecptr++ = acrptr->GroupWrite;
   *vecptr++ = AuthSourceString (acrptr->GroupWrite, acrptr->SourceGroupWrite);

   *vecptr++ = acrptr->GroupRead;
   *vecptr++ = AuthSourceString (acrptr->GroupRead, acrptr->SourceGroupRead);

   *vecptr++ = acrptr->AgentParameter;

   *vecptr++ = acrptr->UserName;
   *vecptr++ = UserCanStringPtr;
   if (rqptr->AuthVmsUserProfileLength)
      *vecptr++ = " (profile)";
   else
      *vecptr++ = "";
   if (acrptr->HttpsOnly)
      *vecptr++ = " (&quot;https:&quot;&nbsp;only)";
   else
      *vecptr++ = "";

   *vecptr++ = acrptr->BasicCount;
   *vecptr++ = acrptr->DigestCount;
   *vecptr++ = DayDateTime(&acrptr->LastAccessBinTime, 20);
   *vecptr++ = acrptr->AccessCount;
   *vecptr++ = acrptr->DataBaseCount;
   if (acrptr->FailureCount < Config.AuthFailureLimit)
   {
      *vecptr++ = acrptr->FailureCount;
      *vecptr++ = "";
      *vecptr++ = 0;
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = 0;
      *vecptr++ = "<FONT COLOR=\"#ff0000\">";
      *vecptr++ = acrptr->FailureCount;
      *vecptr++ = "</FONT>";
   }

   status = sys$faol (&ReportFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   return (SS$_NORMAL);
}

/*****************************************************************************/

